; $Id: asmutils.asm,v 1.6 2001/11/18 03:57:53 bird Exp $
;
; asmutils - assembly utility functions
;
; Copyright (c) 1998-2001 knut st. osmundsen
;
; Project Odin Software License can be found in LICENSE.TXT
;
    .486p

;
; Include files
;
ifdef RING0
    include devsegdf.inc
endif


;
; Exported symbols
;
    public GetCS
    public GetDS
    public GetES
    public GetFS
    public GetGS
    public GetSS
    public DisableInterrupts
    public EnableInterrupts
    public AtomicInc
    public AtomicIncWrap
    public AtomicDec
    public AtomicSet
    public AtomicSetEqual


CODE32 segment
    assume CS:CODE32, DS:FLAT, SS:NOTHING


GetCS proc near
    xor eax,eax
    mov ax,cs
    ret
GetCS endp


GetDS proc near
    xor eax,eax
    mov ax,ds
    ret
GetDS endp


GetES proc near
    xor eax,eax
    mov ax,es
    ret
GetES endp


GetFS proc near
    xor eax,eax
    mov ax,fs
    ret
GetFS endp


GetGS PROC NEAR
    xor eax,eax
    mov ax,gs
    ret
GetGS ENDP


GetSS PROC NEAR
    xor eax,eax
    mov ax,ss
    ret
GetSS ENDP


DisableInterrupts PROC NEAR
    cli
    ret
DisableInterrupts ENDP


EnableInterrupts PROC NEAR
    sti
    ret
EnableInterrupts ENDP


;;
; Performs an SMP safe increment of a give dword number
; @cproto   unsigned long _Optlink AtomicInc(unsigned long *pul)
; @returns  The new value.
; @param    pul     Pointer to unsigned long value to increment.
; @uses     eax, edx, ecx
; @author   knut st. osmundsen (kosmunds@csc.com)
; @remark
AtomicInc proc near
    mov     ecx, eax                    ; ecx = pul
    mov     eax, [ecx]                  ; eax = *pul

retry:
    lea     edx, [eax + 1]              ; edx = *pul + 1
    lock cmpxchg [ecx], edx
    jnz     retry                       ; retry if not changed.

    mov     eax, edx
    ret
AtomicInc endp


;;
; Performs an SMP safe increment of a give dword number with wrapping on some max value.
; @cproto   unsigned long _Optlink AtomicWrap(unsigned long *pul, unsigned long ulMax)
; @returns  The new value.
; @param    pul     Pointer to unsigned long value to decrement.
; @param    ulMax   Maximum value. (not included)
; @uses     eax, edx, ecx
; @author   knut st. osmundsen (kosmunds@csc.com)
; @remark
AtomicIncWrap proc near
    push    ebx
    mov     ebx, edx                    ; ebx = ulMax
    mov     ecx, eax                    ; ecx = pul
    mov     eax, [ecx]                  ; eax = *pul

retry:
    lea     edx, [eax + 1]              ; edx = *pul + 1
    cmp     edx, ebx
    jb      ok
    xor     edx, edx                    ; maximum was reached - wrap.
ok:
    lock cmpxchg [ecx], edx
    jnz     retry                       ; retry if not changed.

    pop     ebx
    mov     eax, edx
    ret
AtomicIncWrap endp


;;
; Performs an SMP safe decrement of a give dword number
; @cproto   unsigned long _Optlink AtomicDec(unsigned long *pul)
; @returns  The new value.
; @param    pul     Pointer to unsigned long value to decrement.
; @uses     eax, edx, ecx
; @author   knut st. osmundsen (kosmunds@csc.com)
; @remark
AtomicDec proc near
    mov     ecx, eax                    ; ecx = pul
    mov     eax, [ecx]                  ; eax = *pul

retry:
    lea     edx, [eax -  1]             ; edx = *pul - 1
    lock cmpxchg [ecx], edx
    jnz     retry                       ; retry if not changed.

    mov     eax, edx
    ret
AtomicDec endp


;;
; Performs an SMP safe set of a give dword number
; @cproto   unsigned long _Optlink AtomicSet(unsigned long *pul, unsigned long ul)
; @returns  The new value.
; @param    pul     Pointer to unsigned long value to decrement.
; @uses     eax, edx, ecx
; @author   knut st. osmundsen (kosmunds@csc.com)
; @remark
AtomicSet proc near
    mov     ecx, eax                    ; ecx = pul
    mov     eax, [ecx]                  ; eax = *pul

retry:
    lock cmpxchg [ecx], edx
    jnz     retry                       ; retry if not changed.

    mov     eax, edx
    ret
AtomicSet endp


;;
; Performs an SMP safe set of a give dword number if it holds a given value.
; @cproto   KBOOL _Optlink AtomicSetEqual(unsigned long *pul, unsigned long ulSet, unsigned long ulCurrent)
; @returns  1 (TRUE) if successfully updated value.
;           0 (FALSE) if current value was not ulCurrent.
; @param    pul         Pointer to unsigned long value to decrement.
; @param    ulSet       The new value to set instead of ulCurrent.
; @param    ulCurrent   The current value of *pul if the operation should succeed.
; @uses     eax, edx, ecx
; @author   knut st. osmundsen (kosmunds@csc.com)
AtomicSetEqual proc near
    xchg    ecx, eax                    ; ecx = pul
                                        ; eax = ulCurrent
                                        ; edx = ulSet
    lock cmpxchg [ecx], edx
    jnz     nok                         ; jump on failure.

    mov     eax, 1
    ret                                 ; success

nok:
    xor     eax, eax
    ret                                 ; failure
AtomicSetEqual endp


CODE32 ends
end
