/* $Id: kAVLGet2.c.h,v 1.2 2001/11/09 07:36:07 bird Exp $
 *
 * kAVLGet - get routine for AVL trees, getting nodes.
 *
 * Copyright (c) 1999-2001 knut st. osmundsen (knut.stange.osmundsen@mynd.no)
 *
 */

#ifndef _kAVLGet2_h_
#define _kAVLGet2_h_


/**
 * Gets a node from the tree (does not remove it!)
 * This is a function ment for checking if a given node really is in the tree.
 * @returns   Pointer to the node holding the given key.
 * @param     ppTree  Pointer to the AVL-tree root node pointer.
 * @param     pNode   Pointer to the node to be found.
 * @sketch
 * @status    completely implemented.
 * @author    knut st. osmundsen
 */
PAVLNODECORE KAVL_FN(Get2)(PPAVLNODECORE ppTree, PAVLNODECORE pNodeGet)
{
    KLOGENTRY2("PAVLNODECORE","PPAVLNODECORE ppTree, PAVLNODECORE pNodeGet", ppTree, pNodeGet);
    AVLKEY                  Key = pNodeGet->Key;
    #ifndef AVL_CMP
    register PAVLNODECORE   pNode = *ppTree;

    while (pNode != NULL && AVL_NE(pNode->Key, Key))
    {
        if (AVL_G(pNode->Key, Key))
            pNode = pNode->pLeft;
        else
            pNode = pNode->pRight;
    }

    #else

    register int            iDiff;
    register PAVLNODECORE   pNode = *ppTree;

    while (pNode != NULL && (iDiff = AVL_CMP(pNode->Key, Key)) != 0)
    {
        if (iDiff > 0)
            pNode = pNode->pLeft;
        else
            pNode = pNode->pRight;
    }

    #endif


    if (pNode == NULL || pNode == pNodeGet)
    {
        KLOGEXIT(pNode);
        return pNode;
    }

    #ifdef KAVL_EQUAL_ALLOWED
    while (pNode->pList != NULL)
    {
        if (pNode->pList == pNodeGet)
        {
            KLOGEXIT(pNode);
            return pNode;
        }
        pNode = pNode->pList;
    }
    #endif

    KLOGEXIT(NULL);
    return NULL;
}



#endif
