/* $Id: kAVLGetWithParent.c.h,v 1.2 2001/11/09 07:36:08 bird Exp $
 *
 * kAVLGetWithParent - get with parent routine for AVL trees.
 *
 * Copyright (c) 1999-2001 knut st. osmundsen (knut.stange.osmundsen@mynd.no)
 *
 */

#ifndef _kAVLGetWithParent_h_
#define _kAVLGetWithParent_h_


/**
 * Gets a node from the tree and its parent node (if any) (does not remove any nodes!)
 * @returns   Pointer to the node holding the given key.
 * @param     ppTree    Pointer to the AVL-tree root node pointer.
 * @param     ppParent  Pointer to a variable which will hold the pointer to the partent node on
 *                      return. When no node is found, this will hold the last searched node.
 * @param     Key       Key value of the node which is to be found.
 * @sketch
 * @status    completely implemented.
 * @author    knut st. osmundsen
 */
PAVLNODECORE    KAVL_FN(GetWithParent)(PPAVLNODECORE ppTree, PPAVLNODECORE ppParent, AVLKEY Key)
{
    KLOGENTRY3("PAVLNODECORE","PPAVLNODECORE ppTree, PPAVLNODECORE ppParent, AVLKEY Key", ppTree, ppParent, Key);
    #ifndef AVL_CMP

    register PAVLNODECORE   pNode = *ppTree;
    register PAVLNODECORE   pParent = NULL;

    while (pNode != NULL && AVL_NE(pNode->Key, Key))
    {
        pParent = pNode;
        if (AVL_G(pNode->Key, Key))
            pNode = pNode->pLeft;
        else
            pNode = pNode->pRight;
    }

    #else

    register PAVLNODECORE   pNode = *ppTree;
    register PAVLNODECORE   pParent = NULL;
    register int            iDiff;

    while (pNode != NULL && (iDiff = AVL_CMP(pNode->Key, Key)) != 0)
    {
        pParent = pNode;
        if (iDiff > 0)
            pNode = pNode->pLeft;
        else
            pNode = pNode->pRight;
    }

    #endif

    *ppParent = pParent;
    KLOGEXIT(pNode);
    return pNode;
}


#endif

