/* $Id: kLog.c,v 1.11 2002/01/09 03:21:58 bird Exp $
 *
 * kLog - Generic Logging and Trace Routines.
 *
 * Copyright (c) 2001 knut st. osmundsen (knut.stange.osmundsen@mynd.no)
 *
 */
#ifndef NOFILEID
static const char szFileId[] = "$Id: kLog.c,v 1.11 2002/01/09 03:21:58 bird Exp $";
#endif


/*******************************************************************************
*   Defined Constants And Macros                                               *
*******************************************************************************/
#define SEQ_MAGIC 0xdeadc0deUL          /* handle magic number. (first dword) */
#define SEQ_FLAGS 0xc0000000UL          /* handle magic flags.  (second dword)*/


#ifdef DEBUG
#   define KLOG_ASSERT(expr)   /* Use with care. Will not allways work... */ \
        (!(expr) \
         ? klog(NULL, NULL, KLOGPOS_EXT, KLOG_TYPE_LOGASSERT, "%s", #expr), INT3() \
         : (void)0, (void)0)
#else
#   define KLOG_ASSERT(expr) \
        (void)0,(void)0
#endif


/*
 * Buffer stuff.
 */
#define KLOG_BUFFERSIZE             0x10000
#define KLOG_BUFFER_ENTRY_SIZE        0x200
#define KLOG_MAX_OVERFLOW           (KLOG_BUFFER_ENTRY_SIZE * 2)
#define KLOG_BUFFER_ENTRIES         ((KLOG_BUFFERSIZE / KLOG_BUFFER_ENTRY_SIZE) - 2) /* - 2 for overflow safety */
#define KLOG_GETBUFFEROFFSET(iBuffer)   ((iBuffer) * KLOG_BUFFER_ENTRY_SIZE)
#define KLOG_GETBUFFERADDRESS(iBuffer) (&achBuffers[KLOG_GETBUFFEROFFSET(iBuffer)])


/*******************************************************************************
*   Header Files                                                               *
*******************************************************************************/
#define INCL_DOSPROCESS
#include <os2.h>
#ifndef RING0
/* only in newer toolkits */
APIRET16  APIENTRY16 Dos16SysTrace(USHORT major,
                                   USHORT cpBuf,
                                   USHORT minor,
                                   ULONG  pf16Buf);
#endif

#include <kKLstdarg.h>
#ifdef RING0
#   define INCL_OS2KRNL_SEM
#   include <OS2Krnl.h>
#   include <OS2Ksem.h>
#   include <kKLsprintf.h>
#   include <kKLprintf.h>
#   include <dev32hlp.h>
#else
#   include <stdio.h>
#endif
#include <kKLasmutils.h>
#include <kTypes.h>
#include <kAssert.h>
#include <kAVL.h>
#define KLOGFUNCTION KLOGFUNCTIONDEF
#include <kLog.h>

#include <string.h>


/*******************************************************************************
*   Global Variables                                                           *
*******************************************************************************/
/*
 * Module tree.
 */
PAVLULNODECORE  pavlModTree;
unsigned long   ulLastModHandle;
#ifdef RING0
KSEMSHR         kshrkLog;
#else

#endif
BOOL            fInited;                /* true when subsystem is fully initiated. */
                                        /* Currenly only used for kshrkLog validations. */

/*
 * Pool of buffers.
 */
KBOOL           afBufferInUse[KLOG_BUFFER_ENTRIES];
extern char     achBuffers[1];
extern ULONG    ulfpBuffers;


/*
 * Default Log Event Types.
 */
struct kLogDefaultTypeInfo
{
    const char *            pszType;
    const char *            pszFormat;
} aDefaultTypes[] =
{
    {"LogInit",             ""},        /* KLOG_TYPE_LOGINIT     */
    {"InitMod",             ""},        /* KLOG_TYPE_INITMOD     */
    {"TermMod",             ""},        /* KLOG_TYPE_TERMMOD     */
    {"Start",               ""},        /* KLOG_TYPE_START       */
    {"Entry",               ""},        /* KLOG_TYPE_ENTRY       */
    {"Exit",                ""},        /* KLOG_TYPE_EXIT        */
    {"Stop",                ""},        /* KLOG_TYPE_STOP        */
    {"Printf",            NULL},        /* KLOG_TYPE_PRINTF     - expects an format string with args. */
    {"Assert",            "%s"},        /* KLOG_TYPE_ASSERT     - expects an expression string. */
    {"kLogAssert",        NULL},        /* KLOG_TYPE_LOGASSERT  - expects an format string with args. */
};


/*
 * Default struct member info.
 */

#ifdef INCL_OS2KRNL_LDR
KLOGMEMBERINFO  aklogMemb_MTE[] =
{
    {"modname",     "%.8s",  KLOG_TYPEDEF_ARRAY | KSIZEOF(MTE, mte_modname), KOFFSETOF(MTE, mte_modname)},
    {"handle",      "0x%04x",   KSIZEOF(MTE, mte_handle),  KOFFSETOF(MTE, mte_handle)},
    {"flags1",      "0x%08x",   KSIZEOF(MTE, mte_flags),   KOFFSETOF(MTE, mte_flags)},
    {"flags2",      "0x%04x",   KSIZEOF(MTE, mte_flags2),  KOFFSETOF(MTE, mte_flags2)},
};
#endif

/*
 * Default typedef info.
 */
PAVLSTRNODECORE pavlklogTypedefs;
KLOGTYPEDEFINFO aklogTypedefs[] =
{
    {"PSCZ",            "%.80s", KLOG_TYPEDEF_PTR | 4,  0,                          NULL},
    {"PSZ",             "%.80s", KLOG_TYPEDEF_PTR | 4,  0,                          NULL},
    {"char *",          "%.80s", KLOG_TYPEDEF_PTR | 4,  0,                          NULL},
    {"const char *",    "%.80s", KLOG_TYPEDEF_PTR | 4,  0,                          NULL},
    {"char const *",    "%.80s", KLOG_TYPEDEF_PTR | 4,  0,                          NULL},
    {"char",            "0x%02x",        sizeof(char),  0,                          NULL},
    {"signed char",     "0x%02x",        sizeof(char),  0,                          NULL},
    {"unsigned char",   "0x%02x",        sizeof(char),  0,                          NULL},
    {"short",           "%d",           sizeof(short),  0,                          NULL},
    {"SHORT",           "%d",           sizeof(short),  0,                          NULL},
    {"unsigned short",  "0x%04x",       sizeof(short),  0,                          NULL},
    {"USHORT",          "0x%04x",       sizeof(short),  0,                          NULL},
    {"signed",          "%d",             sizeof(int),  0,                          NULL},
    {"signed int",      "%d",             sizeof(int),  0,                          NULL},
    {"int",             "%d",             sizeof(int),  0,                          NULL},
    {"INT",             "%d",             sizeof(int),  0,                          NULL},
    {"unsigned",        "0x%08x",         sizeof(int),  0,                          NULL},
    {"unsigned int",    "0x%08x",         sizeof(int),  0,                          NULL},
    {"UINT",            "0x%08x",         sizeof(int),  0,                          NULL},
    {"long",            "%ld",           sizeof(long),  0,                          NULL},
    {"LONG",            "%ld",           sizeof(long),  0,                          NULL},
    {"signed long",     "%ld",           sizeof(long),  0,                          NULL},
    {"unsigned long",   "0x%08x",        sizeof(long),  0,                          NULL},
    {"ULONG",           "0x%08x",        sizeof(long),  0,                          NULL},
#ifdef INCL_OS2KRNL_LDR
    {"PMTE",            "%.80s", KLOG_TYPEDEF_PTR | 4,  KCENTRIES(aklogMemb_MTE),   &aklogMemb_MTE[0]},
    {"MTE",             "%.80s",          sizeof(MTE),  KCENTRIES(aklogMemb_MTE),   &aklogMemb_MTE[0]},
#endif
    {"", "",  0, 0, NULL}
};


/*
 * Settings.
 */
#ifndef fkLogCurrent
unsigned short  fkLogCurrent = KLOG_FLAGS_ENABLED;
#endif

/*
 * Global info.
 */
ULONG   ulGlobalLastSeq;

/*******************************************************************************
*   Special AVL Functions                                                      *
*******************************************************************************/
PAVLULNODECORE      KLIBCALL AVLULNLRemove(PPAVLULNODECORE ppTree, AVLULKEY Key);
KBOOL               KLIBCALL AVLULNLInsert(PPAVLULNODECORE ppTree, PAVLULNODECORE pNode);
PAVLULNODECORE      KLIBCALL AVLULNLGet(PPAVLULNODECORE ppTree, AVLULKEY Key);
PAVLSTRNODECORE     KLIBCALL AVLStrNLRemove(PPAVLSTRNODECORE ppTree, AVLSTRKEY Key);
KBOOL               KLIBCALL AVLStrNLInsert(PPAVLSTRNODECORE ppTree, PAVLSTRNODECORE pNode);
PAVLSTRNODECORE     KLIBCALL AVLStrNLGet(PPAVLSTRNODECORE ppTree, AVLSTRKEY Key);


/*******************************************************************************
*   Internal Functions                                                         *
*******************************************************************************/
INLINE PKLOGMODDATA klogGetMod(HKLOGMOD hLogMod);
PHKLOGSEQ _System   klogGetSeq(int iDummy);
INLINE int          klogGetBuffer(void);
INLINE void         klogFreeBuffer(int iBuffer);
void                klogEntry(PHKLOGSEQ phSeq, PKLOGMODDATA pMod, KLOGPOS_DECL, int iType, const char *pszFuncProto, va_list args);
void                klogExit( PHKLOGSEQ phSeq, PKLOGMODDATA pMod, KLOGPOS_DECL, int iType, const char *pszReturnType, va_list retarg);
void                klog(     PHKLOGSEQ phSeq, PKLOGMODDATA pMod, KLOGPOS_DECL, int iType, ...);
void                klogv(    PHKLOGSEQ phSeq, PKLOGMODDATA pMod, KLOGPOS_DECL, int iType, va_list args);
unsigned short      klogGetFlags(PKLOGMODDATA pMod, const char *pszFile, int iType);
char *              klogInterpretType(PKLOGMODDATA pMod, char *psz, unsigned short fFlags, const char *pachType, int cchType, va_list *parg);

void                klogBufferPut(unsigned short usFlags, int iBuffer, int cchBuffer);


/**
 * Resolves a module handle into a module data structure.
 * @returns Pointer to module structure.
 * @param   hLogMod     Module Log handle.
 * @author  knut st. osmundsen (kosmunds@csc.com)
 */
INLINE PKLOGMODDATA klogGetMod(HKLOGMOD hLogMod)
{
    char *pch;
    if (!fInited)
        return NULL;
#ifdef RING0
    if (KSEMRequestShared(&kshrkLog, KSEM_INDEFINITE_WAIT))
        return NULL;
    pch = (char*)(void*)AVLULNLGet(&pavlModTree, (AVLULKEY)hLogMod);
    KSEMRelease((HKSEM)(void*)&kshrkLog);
#else
    pch = (char*)(void*)AVLULNLGet(&pavlModTree, (AVLULKEY)hLogMod);
#endif
    KLOG_ASSERT(pch || !hLogMod);
    return pch ? (PKLOGMODDATA)(void*)(pch - KOFFSETOF(KLOGMODDATA, avlnodecore)) : NULL;
}


/**
 * Finds the sequence number from the stack.
 * @returns Pointer to sequence handle if found.
 *          NULL if not found.
 * @param   iDummy  Dummy parameter just to determin stack address.
 * @sketch
 *          Get current stack pointer.
 *          Get pointer to top of stack.
 *          Scan for sequence magic number from current pointer to top of stack.
 *
 * @status
 * @author  knut st. osmundsen (kosmunds@csc.com)
 * @remark
 */
PHKLOGSEQ klogGetSeq(int iDummy)
{
    char *  pchStack = (char*)(void*)SSToDS(&iDummy);
    char *  pchStackTop;

    /*
     * Determin stack top.
     */
    #ifdef RING0
    pchStackTop = (char*)KALIGNUP(PAGESIZE, (unsigned)pchStack);
    #else
    PPIB    pPib;
    PTIB    pTib;
    DosGetInfoBlocks(SSToDS(&pTib), SSToDS(&pPib));
    pchStackTop = pTib->tib_pstacklimit;
    #endif

    /*
     * Scan the stack for the sequence tag.
     * Scan rules:
     *      Aligned on 8 byte boundrary.
     *      Starts with magic dword.
     *      Second dword contains the magic flags.
     */
    pchStack = (char*)KALIGNUP(8, pchStack);
    while (pchStack < pchStackTop)
    {
        if (    ((PHKLOGSEQ)(void*)pchStack)->data.ulMagic == SEQ_MAGIC
            &&  ((((PHKLOGSEQ)(void*)pchStack)->data.ulSeq & SEQ_FLAGS) == SEQ_FLAGS)
            )
            return (PHKLOGSEQ)(void*)pchStack;
        pchStack += 8;
    }

    /*
     * not found
     */
    return NULL;
}


/**
 * Allocates a buffer from the pool.
 * @returns The index of the allocated buffer.
 * @remark  Will rather panic than fail.
 */
INLINE int          klogGetBuffer(void)
{
    int j;
    int i;
    int aiDist[8] = {0,4,2,6,1,3,5,7};

    do
    {
        /*
         * Look for a free buffer.
         * (We try to distribute the used entries in order to somehow allow overflows.)
         */
        for (j = 0; j < 8; j++)
        {
            for (i = aiDist[j]; i < KCENTRIES(afBufferInUse); i += 8)
            {
                if (AtomicSetEqual(&afBufferInUse[i], TRUE, FALSE))
                    return i;
            }
        }

        #ifdef DEBUG
        /* panic */
        KLOG_ASSERT(!j);//(0); shutup compilers.
        #endif

        /* Yield the CPU in order to make one of the 64 thread blocked with a buffer
         * from the pool able to free the buffer.
         * THIS SHOULD NOT HAPPEN.
         */
        #ifdef RING0
        D32Hlp_Yield();
        #else
        DosSleep(0);
        #endif
    } while (i);    /* Use i to shutup the compiler (while(TRUE) is what should be here). */
    return -1;
}


/**
 * Frees a buffer in use.
 * @param   iBuffer     Buffer index of the buffer to be freed.
 */
INLINE void         klogFreeBuffer(int iBuffer)
{
    KLOG_ASSERT(iBuffer < KCENTRIES(afBufferInUse) && afBufferInUse[iBuffer]);
    AtomicSet(&afBufferInUse[iBuffer], FALSE);
}


/**
 * Formats the common parts of all events.
 * @returns Pointer to the end of what we've written into pszBuffer.
 * @param   pszBuffer       Output buffer.
 * @param   phSeq           Pointer to sequence handle. NULL is allowed, but if so global event is assumed.
 * @param   pMod            Module data.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   iType           Logging event type.
 */
char *klogFormatCommon(char *pszBuffer, PHKLOGSEQ phSeq, PKLOGMODDATA pMod, KLOGPOS_DECL, int iType)
{
    const char *psz;
    int         cch;

    if (*pszFile == '$' && pszFile[1] == 'I' && pszFile[2] == 'd')
    {
        cch = 0;
        psz = pszFile + 5;
        //while (psz[cch] != '\0' && *(const unsigned short *)(psz + cch) != (unsigned short)(',' + ('v' >> 8)))
        while (psz[cch] != '\0' && (psz[cch] != ',' || psz[cch + 1] != 'v'))
            cch++;
    }
    else
    {
        cch = 0;
        psz = pszFile + strlen(pszFile);

        /* find start of filename. */
        while (psz >= pszFile && *psz != '\\' && *psz != '/' && *psz != ':')
            psz--, cch++;
        psz++;
    }

    /* write module, sequence, filename and line */
    pszBuffer += sprintf(pszBuffer, "%s;%d;%.*s(%d);",
                         pMod ? pMod->pszModuleName             : "<global>",
                         phSeq ? phSeq->data.ulSeq & ~SEQ_FLAGS : 0,
                         cch, psz,
                         iLine);


    #if 1                               /* may be disabled if no pszFunction... */
    /* Find start of function
     * Rules:
     *      The first word which has an parenthesis 'postfixed'.
     *      The entire string if no parenthesis found.
     */
    psz = strchr(pszFunction, '(');
    if (psz)
    {
        do
        {
            if (psz[-1] != ' ')
            {
                register char ch;
                const char *pszFn = psz - 1;
                while (     (ch = *pszFn) == '_'
                       ||   (ch >= 'a' && ch <= 'z')
                       ||   (ch >= 'A' && ch <= 'Z')
                       ||   (ch >= '0' && ch <= '9')
                       )
                    pszFn--;
                strncat(pszBuffer, pszFn + 1, psz - pszFn - 2);
                break;
            }
            psz = strchr(psz + 1, '(');
            KLOG_ASSERT(psz && "unsupported __FUNCTION__ format");
        } while (psz);
    }
    else
        strcpy(pszBuffer, pszFunction);
    pszBuffer += strlen(pszBuffer);
    strcpy(pszBuffer, "()");
    pszBuffer += 2;
    #endif

    /*
     * Add type.
     */
    pszBuffer += sprintf(pszBuffer, ";%x;%s: ", iType,
                         iType < KCENTRIES(aDefaultTypes)
                         ? aDefaultTypes[iType].pszType
                         : iType >= KLOG_TYPE_USER && pMod && pMod->cUserTypes + KLOG_TYPE_USER > iType
                           ? pMod->paUserTypes[iType].pszType
                           : "????????"
                         );


    return pszBuffer;
}


void klogEntry(PHKLOGSEQ phSeq, PKLOGMODDATA pMod, KLOGPOS_DECL, int iType, const char *pszFuncProto, va_list args)
{
    char *          pszBuffer;
    char *          pszCur;
    int             iBuffer;
    unsigned short  fFlags;

    /*
     * Get current settings.
     */
    fFlags = klogGetFlags(pMod, pszFile, iType);


    /*
     * Check if there is actually anything to do.
     */
    if (!fFlags)
        return;


    /*
     * Allocate buffer.
     */
    iBuffer = klogGetBuffer();
    pszBuffer = KLOG_GETBUFFERADDRESS(iBuffer);


    /*
     * Log common stuff.
     */
    pszCur = klogFormatCommon(pszBuffer, phSeq, pMod, KLOGPOS_INT, iType);


    /*
     * Parameters stuff.
     */
    while (pszFuncProto && *pszFuncProto != '\0')
    {
        int         cParatheses = 0;
        const char *pszType = pszFuncProto; /* assumes type first*/
        int         cchType;
        const char *pszName;
        int         cchName;

        /* find end of parameter */
        while ((*pszFuncProto != ',' && *pszFuncProto != '\0') || cParatheses > 0)
        {
            if (*pszFuncProto == '(')
                cParatheses++;
            else if (*pszFuncProto == ')')
                cParatheses--;
            pszFuncProto++;
        }

        /* Find arg name:
         *      *pszFuncProto is now either '\0' or ','.
         *      So, we have to wind back to the last blank.
         *  We doesn't parse function arguments correctly.
         */
        for (pszName = pszFuncProto, cchName = -1; pszName > pszType && *pszName != ' '; pszName--)
        {
            if (*pszName == ')' || *pszName == '(')
            {
                pszType = "void * <pfn>";
                cchType = 6;
                pszName = pszType + 7;
                cchName = 5;
                break;
            }
        }
        if (cchName < 0)                /* indicates no "<pfn>". */
        {
            if (pszName > pszType)
            {   /* *pszName == ' ' */
                cchType = pszName - pszType; /* assumes single space! */
                cchName = pszFuncProto - ++pszName;
            }
            else
            {   /* bad argument somehow: can't continue. */
                break;
            }
        }


        /*
         * Lookup type if possible.
         */
        pszCur += sprintf(pszCur, "%.*s=", cchName, pszName);
        pszCur = klogInterpretType(pMod, pszCur, fFlags, pszType, cchType, SSToDS(&args));

        /* Next:
         *  skip comma and blanks
         */
        while (*pszFuncProto == ',' || *pszFuncProto == ' ')
            pszFuncProto++;

        /*
         * Is this the end?
         */
        if (!*pszFuncProto)
            break;
        *pszCur++ = ',';
        *pszCur++ = ' ';
    }

    /*
     * Put into the log buffer.
     */
    klogBufferPut(fFlags, iBuffer, pszCur - pszBuffer);


    /*
     * Free buffer.
     */
    klogFreeBuffer(iBuffer);
}


/**
 *
 */
void klogExit( PHKLOGSEQ phSeq, PKLOGMODDATA pMod, KLOGPOS_DECL, int iType, const char *pszReturnType, va_list retarg)
{
    char *          pszBuffer;
    char *          pszCur;
    int             iBuffer;
    unsigned short  fFlags;

    /*
     * Get current settings.
     */
    fFlags = klogGetFlags(pMod, pszFile, iType);


    /*
     * Check if there is actually anything to do.
     */
    if (!fFlags)
        return;


    /*
     * Allocate buffer.
     */
    iBuffer = klogGetBuffer();
    pszBuffer = KLOG_GETBUFFERADDRESS(iBuffer);


    /*
     * Log common stuff.
     */
    pszCur = klogFormatCommon(pszBuffer, phSeq, pMod, KLOGPOS_INT, iType);


    /*
     * Type specific stuff.
     */
    if (!pszReturnType)
    {
        unsigned uReturn = va_arg(retarg, unsigned);
        pszCur += sprintf(pszCur, "returns 0x%08x (%d)", uReturn, uReturn);
    }
    else
    {
        strcpy(pszCur, "returns ");     pszCur += 8;
        pszCur = klogInterpretType(pMod, pszCur, fFlags, pszReturnType, strlen(pszReturnType), SSToDS(&retarg));
    }


    /*
     * Put into the log buffer.
     */
    klogBufferPut(fFlags, iBuffer, pszCur - pszBuffer);


    /*
     * Free buffer.
     */
    klogFreeBuffer(iBuffer);
}


/**
 * Internal function for doing logging of events with a variable number of parameters.
 * @param   phSeq           Pointer to sequence handle. NULL is allowed, but if so global event is assumed.
 * @param   iType           Logging event type.
 * @param   pMod            Module data.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   ...             Extra parameters specific to the type.
 */
void klog(PHKLOGSEQ phSeq, PKLOGMODDATA pMod, KLOGPOS_DECL, int iType, ...)
{
    va_list     args;

    va_start(args, iType);
    klogv(phSeq, pMod, KLOGPOS_INT, iType, args);
    va_end(args);
}


/**
 * Internal function for doing logging of events with a variable number of parameters.
 * @param   phSeq           Pointer to sequence handle. NULL is allowed, but if so global event is assumed.
 * @param   iType           Logging event type.
 * @param   pMod            Module data.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   args            Extra parameters specific to the type.
 */
void klogv(PHKLOGSEQ phSeq, PKLOGMODDATA pMod, KLOGPOS_DECL, int iType, va_list args)
{
    char *          pszBuffer;
    char *          pszCur;
    int             iBuffer;
    unsigned short  fFlags;

    /*
     * Get current settings.
     */
    fFlags = klogGetFlags(pMod, pszFile, iType);


    /*
     * Check if there is actually anything to do.
     */
    if (!fFlags)
        return;


    /*
     * Allocate buffer.
     */
    iBuffer = klogGetBuffer();
    pszBuffer = KLOG_GETBUFFERADDRESS(iBuffer);


    /*
     * Log common stuff.
     */
    pszCur = klogFormatCommon(pszBuffer, phSeq, pMod, KLOGPOS_INT, iType);


    /*
     * Type specific stuff.
     */
    if (iType < KLOG_TYPE_USER)
    {
        if (KCENTRIES(aDefaultTypes) > iType)
        {
            const char *pszFormat = aDefaultTypes[iType].pszFormat
                ? aDefaultTypes[iType].pszFormat : va_arg(args, const char *);
            pszCur += vsprintf(pszCur, pszFormat, args);
        }
        else
            pszCur += sprintf(pszCur, "Unknown log event type %d", iType);
    }
    else if (pMod)
    {
        iType -= KLOG_TYPE_USER;
        if (pMod->cUserTypes > iType)
        {
            const char *pszFormat = pMod->paUserTypes[iType].pszFormat
                ? pMod->paUserTypes[iType].pszFormat : va_arg(args, const char *);
            pszCur += vsprintf(pszCur, pszFormat, args);
        }
        else
            pszCur += sprintf(pszCur, "Unknown user log event type %d", iType + KLOG_TYPE_USER);
    }
    else
        pszCur += sprintf(pszCur, "No valid module pointer. (User log event type %d)", iType);


    /*
     * Put into the log buffer.
     */
    klogBufferPut(fFlags, iBuffer, pszCur - pszBuffer);


    /*
     * Free buffer.
     */
    klogFreeBuffer(iBuffer);
}


/**
 * Gets the flags valid for this log event.
 * @returns Flags.
 * @param   pMod        Pointer to module handle.
 *                      NULL if global log event.
 * @param   pszFile     Name of the source file where the event occured.
 * @param   iType       Log event Type.
 * @author  knut st. osmundsen (kosmunds@csc.com)
 */
unsigned short klogGetFlags(PKLOGMODDATA pMod, const char *pszFile, int iType)
{
    unsigned short  fFlags;

    /*
     * Global Settings First.
     */
    fFlags = fkLogCurrent;
    if (fFlags && pMod)
    {
        /*
         * Module Level.
         */
        if (pMod->fCurrent != KLOG_FLAGS_AS_ABOVE)
            fFlags = pMod->fCurrent;
        if (fFlags)
        {
            /*
             * Type Level.
             */
            if (iType < KLOG_TYPE_USER)
            {
                if (    KCENTRIES(aDefaultTypes) > iType
                    &&  pMod->paDefTypes[iType].fCurrent != KLOG_FLAGS_AS_ABOVE
                    )
                    fFlags = pMod->paDefTypes[iType].fCurrent;
            }
            else
            {
                iType -= KLOG_TYPE_USER;
                if (    pMod->cUserTypes > iType
                    &&  pMod->paUserTypes[iType].fCurrent != KLOG_FLAGS_AS_ABOVE
                    )
                    fFlags = pMod->paUserTypes[iType].fCurrent;
            }

            if (fFlags && pMod->cFiles)
            {
                /*
                 * Then File Level.
                 */
                PKLOGFILEINFO pFile = (PKLOGFILEINFO)(void*)AVLStrNLGet(&pMod->pavlFiles, pszFile);
                if (pFile && pFile->fCurrent != KLOG_FLAGS_AS_ABOVE)
                    fFlags = pFile->fCurrent;
            } /* file */
        } /* type */
    } /* module */

    return fFlags;
}

/**
 * Looks up..
 */
char *klogInterpretType(PKLOGMODDATA pMod, char *psz, unsigned short fFlags, const char *pachType, int cchType, va_list *parg)
{
    PKLOGTYPEDEFINFO    pTypedef = NULL;

    /*
     * Look for pszType in pMod first.
     */
    if (pachType)
    {
        char *pszType;

        /* Make zero terminate copy of the type. */
        pszType = SSToDS(alloca(cchType + 1));
        memcpy(pszType, pachType, cchType);
        pszType[cchType] = '\0';

        /* Do the lookups. */
        if (pMod && pMod->cTypedefs > 0)
            pTypedef = (PKLOGTYPEDEFINFO)(void*)AVLStrNLGet(&pMod->pavlTypedefs, pszType);
        if (!pTypedef)
            pTypedef = (PKLOGTYPEDEFINFO)(void*)AVLStrNLGet(&pavlklogTypedefs, pszType);
    }

    /*
     * Found it?
     */
    if (pTypedef)
    {
        if (pTypedef->cMembers)
        {   /*
             * Structure with members.
             */
            char *pch;
            if (pTypedef->cb & KLOG_TYPEDEF_PTR)
                pch = va_arg(*parg, char *);
            else
            #ifdef __IBMC__
                pch = *(char**)parg;  /* this is not portable... but it won't happen so often.*/
            #else
                pch = NULL;
            #endif

            if (    pch >= (char*)0x10000
            #ifdef RING0
                &&  pch < (char*)0xc0000000)
            #else
                &&  pch < (char*)0xfffe0000)
            #endif
            {
                int i;
                for (i = 0; i < pTypedef->cMembers; i++)
                {
                    psz += sprintf(psz, "%s:", pTypedef->paMembers[i].pszName);

                    if (pTypedef->paMembers[i].cb & KLOG_TYPEDEF_ARRAY)
                    {  /*
                        * This is an array of some kind.
                        */
                        psz += sprintf(psz, pTypedef->paMembers[i].pszFormat,
                                       &pch[pTypedef->paMembers[i].offset]);
                    }
                    else
                    {  /*
                        * This is value or pointer.
                        */
                        KSIZE   val;
                        switch (pTypedef->paMembers[i].cb & KLOG_TYPEDEF_SIZE)
                        {
                            case 1: val = pch[pTypedef->paMembers[i].offset]; break;
                            case 2: val = *(unsigned short*)(void*)(pch + pTypedef->paMembers[i].offset); break;
                            case 4: val = *(unsigned long*)(void*)(pch + pTypedef->paMembers[i].offset); break;
                            default:
                                INT3();
                        }
                        psz += sprintf(psz, pTypedef->paMembers[i].pszFormat, val);
                    }
                }
                *psz++ = '}';
            }
            else
                psz += sprintf(psz, "invalid pointer}");
        }
        else
        {   /*
             * Simple type (no struct). Use vsprintf to format it.
             */
            psz += vsprintf(psz, pTypedef->pszFormat, *parg);
        }

        /*
         * Advance *parg (a bit tricky).
         */
        *parg = (va_list)((KSIZE)*parg  + KALIGNUP(sizeof(KSIZE), pTypedef->cb & KLOG_TYPEDEF_SIZE));
    }
    else
    {
        /*
         * Didn't find the type, check for void.
         */
        if (pachType && cchType == 4 && !strnicmp(pachType, "void", 4))
        {
            strcpy(psz, "<void>");
            psz += 7;
        }
        else
        {
            if (*parg)                  /* can't be too careful. */
            {
                unsigned uValue = va_arg(*parg, unsigned);
                psz += sprintf(psz, "0x%08x", uValue);
            }
            else
                psz += sprintf(psz, "???");
        }
    }

    return psz;
}


/**
 * Puts log/trace data into some buffer system so we can collect it later..
 * @param   usFlags     Logging flags. (determins what we should do).
 * @param   iBuffer     The index of the buffer we're to put.
 * @param   cchBuffer   Size of the data we're to put.
 */
void klogBufferPut(unsigned short usFlags, int iBuffer, int cchBuffer)
{
    #ifdef DEBUG
    /*
     * Check for overflow.
     */
    if (cchBuffer >= KLOG_MAX_OVERFLOW)
    {
        klog(NULL, NULL, KLOGPOS_EXT, KLOG_TYPE_LOGASSERT,
             "overflowed buffer %d with %d bytes",
             iBuffer,
             cchBuffer - KLOG_BUFFER_ENTRY_SIZE);
    }
    #endif

    if (usFlags & KLOG_FLAGS_ENABLED)
    {
        /*
         * Skip any new line at end of buffer.
         */
        int cch = cchBuffer;
        char *psz = KLOG_GETBUFFERADDRESS(iBuffer);
        if (cch > 0 && psz[cch - 1] == '\n' || psz[cch - 1] == '\r')
        {
            cch--;
            if (cch > 1 && psz[cch - 2] == '\n' || psz[cch - 2] == '\r')
                cch--;
        }
        #ifdef RING0
        D32Hlp_RAS(
        #else
        Dos16SysTrace(
        #endif
            KLOG_MAJOR,
            cch,
            KLOG_MINOR,
            ulfpBuffers + KLOG_GETBUFFEROFFSET(iBuffer));
    }

    #if KLOG_FLAGS_COMOUTPUT > 0
    if (usFlags & KLOG_FLAGS_COMOUTPUT)
        printf("%.*s", cchBuffer, KLOG_GETBUFFEROFFSET(iBuffer));
    #endif
}




/**
 * Initiates the loging system.
 * @returns 0 on success.
 *          -1 on error?
 * @param   Size of the static logging buffer.
 */
int         kLogInit(unsigned long cbBufferSize)
{
    int i;

    /*if (!kLogInitBuffer(cbBufferSize))
        return FALSE;*/

    #ifdef RING0
    KSEMInit((PKSEM)(void*)&kshrkLog, KSEM_SHARED, 0);
    #else
    #endif
    pavlModTree = NULL;
    ulLastModHandle = 0;

    /*
     * Init the type avl tree.
     */
    pavlklogTypedefs = NULL;
    for (i = 0; i < KCENTRIES(aklogTypedefs); i++)
        AVLStrNLInsert(&pavlklogTypedefs, &aklogTypedefs[i].u1.avlnodecore);

    fInited = TRUE;

    KNOREF(cbBufferSize);
    kLog(0, KLOGPOS_EXT, 0);
    return 0;
}


/**
 * Initiates logging for a module.
 * @returns Logging handle for the module.
 * @param   pMod            Pointer to the log data.
 * @param   pszFile         Filename where we're called from.
 * @param   iLine           Line in that file.
 * @param   pszFunction     The function name.
 */
HKLOGMOD kLogInitMod(PKLOGMODDATA pMod, KLOGPOS_DECL)
{
    #ifdef RING0
    ULONG rc;
    #endif
    int i;

    /*
     * Validate input.
     */
    if (    !pMod->pszModuleName || !*pMod->pszModuleName
        ||  !pMod->paDefTypes    || pMod->cDefTypes < KLOG_TYPE_LAST_REQ
        )
    {
        KLOG_ASSERT(pMod->pszModuleName && *pMod->pszModuleName
                    && pMod->paDefTypes && pMod->cDefTypes >= KLOG_TYPE_LAST_REQ);
        return 0;
    }


    #ifdef DEBUG
    /*
     * Validate the typedef structures.
     */
    for (i = 0; i < pMod->cTypedefs; i++)
    {
        int j;
        KLOG_ASSERT((pMod->paTypedefs[i].cb & KLOG_TYPEDEF_SIZE) > 0);
        KLOG_ASSERT(  !(pMod->paTypedefs[i].cb & KLOG_TYPEDEF_PTR)
                    || (pMod->paTypedefs[i].cb & KLOG_TYPEDEF_SIZE) == 4);
        KLOG_ASSERT(   pMod->paTypedefs[i].pszFormat
                    && *pMod->paTypedefs[i].pszFormat
                    && strchr(pMod->paTypedefs[i].pszFormat, '%'));
        KLOG_ASSERT(   pMod->paTypedefs[i].u1.pszType
                    && *pMod->paTypedefs[i].u1.pszType
                    && !strstr(pMod->paTypedefs[i].u1.pszType, "  "));
        KLOG_ASSERT(   (!pMod->paTypedefs[i].cMembers && !pMod->paTypedefs[i].paMembers)
                    || (pMod->paTypedefs[i].cMembers > 0 && pMod->paTypedefs[i].paMembers));

        for (j = 0; j < pMod->paTypedefs[i].cMembers; j++)
        {
            KLOG_ASSERT((pMod->paTypedefs[i].paMembers[j].cb & KLOG_TYPEDEF_SIZE) > 0);
            KLOG_ASSERT(  !(pMod->paTypedefs[i].paMembers[j].cb & KLOG_TYPEDEF_PTR)
                        || (pMod->paTypedefs[i].paMembers[j].cb & KLOG_TYPEDEF_SIZE) == 4);
            KLOG_ASSERT(   pMod->paTypedefs[i].paMembers[j].pszFormat
                        && *pMod->paTypedefs[i].paMembers[j].pszFormat
                        && strchr(pMod->paTypedefs[i].paMembers[j].pszFormat, '%'));
            KLOG_ASSERT(    pMod->paTypedefs[i].paMembers[j].pszName
                        && *pMod->paTypedefs[i].paMembers[j].pszName);
        }
    }

    /*
     * Validate the defaults types.
     */
    for (i = 0; i < pMod->cDefTypes; i++)
    {
        KLOG_ASSERT(   (pMod->paDefTypes[i].fCurrent & ~KLOG_FLAGS_MASK) == 0
                    ||  pMod->paDefTypes[i].fCurrent == KLOG_FLAGS_AS_ABOVE);
        KLOG_ASSERT(   (pMod->paDefTypes[i].fDefault & ~KLOG_FLAGS_MASK) == 0
                    ||  pMod->paDefTypes[i].fDefault == KLOG_FLAGS_AS_ABOVE);
    }

    /*
     * Validate the user types.
     */
    for (i = 0; i < pMod->cUserTypes; i++)
    {
        KLOG_ASSERT(   (pMod->paUserTypes[i].fCurrent & ~KLOG_FLAGS_MASK) == 0
                    ||  pMod->paUserTypes[i].fCurrent == KLOG_FLAGS_AS_ABOVE);
        KLOG_ASSERT(   (pMod->paUserTypes[i].fDefault & ~KLOG_FLAGS_MASK) == 0
                    ||  pMod->paUserTypes[i].fDefault == KLOG_FLAGS_AS_ABOVE);
        KLOG_ASSERT(    pMod->paUserTypes[i].pszType
                    &&  *pMod->paUserTypes[i].pszType
                    &&  !strstr(pMod->paUserTypes[i].pszType, "  "));
    }

    /*
     * Validate the file info.
     */
    for (i = 0; i < pMod->cFiles; i++)
    {
        KLOG_ASSERT(   (pMod->paFiles[i].fCurrent & ~KLOG_FLAGS_MASK) == 0
                    ||  pMod->paFiles[i].fCurrent == KLOG_FLAGS_AS_ABOVE);
        KLOG_ASSERT(   (pMod->paFiles[i].fDefault & ~KLOG_FLAGS_MASK) == 0
                    ||  pMod->paFiles[i].fDefault == KLOG_FLAGS_AS_ABOVE);
        KLOG_ASSERT(   !strchr(pMod->paFiles[i].u1.pszFile, ':')
                    && !strchr(pMod->paFiles[i].u1.pszFile, '/')
                    && !strchr(pMod->paFiles[i].u1.pszFile, '\\'));
    }
    #endif

    /*
     * Init module data.
     */
    pMod->ulLastSeq = 0;
    memset(&pMod->avlnodecore, 0, sizeof(pMod->avlnodecore));
    pMod->avlnodecore.Key = AtomicInc(&ulLastModHandle);


    /*
     * Init the type avl tree.
     */
    pMod->pavlTypedefs = NULL;
    for (i = 0; i < pMod->cUserTypes; i++)
        AVLStrNLInsert(&pMod->pavlTypedefs, &pMod->paTypedefs[i].u1.avlnodecore);


    /*
     * Init the type avl tree.
     */
    pMod->pavlFiles = NULL;
    for (i = 0; i < pMod->cFiles; i++)
        AVLStrNLInsert(&pMod->pavlFiles, &pMod->paFiles[i].u1.avlnodecore);


    /*
     * Add the module to the avl tree.
     */
    #ifdef RING0
    rc = KSEMRequestExclusive(&kshrkLog, KSEM_INDEFINITE_WAIT);
    if (rc)
    {
        klog(NULL, pMod, KLOGPOS_EXT, KLOG_TYPE_LOGASSERT,
             "KSEMRequestExclusive failed, rc=%d", rc);
        return NULLHANDLE;
    }
    AVLULNLInsert(&pavlModTree, &pMod->avlnodecore);
    KSEMRelease((HKSEM)(void*)&kshrkLog);
    #else
    AVLULNLInsert(&pavlModTree, &pMod->avlnodecore);
    #endif

    /*
     * Log the event.
     */
    klogv(NULL, pMod, KLOGPOS_INT, KLOG_TYPE_INITMOD, NULL);

    /*
     * Return successfully.
     */
    return pMod->avlnodecore.Key;
}

/**
 * Terminates logging for a given module.
 * @param   hLogMod         Handle of the module to terminate logging for.
 * @param   pszFile         Filename where we're called from.
 * @param   iLine           Line in that file.
 * @param   pszFunction     The function name.
 */
void kLogTermMod(HKLOGMOD hLogMod, KLOGPOS_DECL)
{
    PKLOGMODDATA    pMod;
    #ifdef RING0
    ULONG   rc;
    #endif

    /*
     * Get module data from handle.
     */
    if ((pMod = klogGetMod(hLogMod)) == NULL)
        return;

    /*
     * Get exclusive access to avl tree.
     * Remove the node.
     * Release shared sem.
     */
    #ifdef RING0
    rc = KSEMRequestExclusive(&kshrkLog, KSEM_INDEFINITE_WAIT);
    if (rc)
    {
        klog(NULL, pMod, KLOGPOS_EXT, KLOG_TYPE_LOGASSERT,
             "KSEMRequestExclusive failed, rc=%d", rc);
        return;
    }
    #else
    #endif

    klogv(NULL, pMod, KLOGPOS_INT, KLOG_TYPE_TERMMOD, NULL);

    pMod = (PKLOGMODDATA)(void*)AVLULNLRemove(&pavlModTree, pMod->avlnodecore.Key);

    #ifdef RING0
    KSEMRelease((HKSEM)(void*)&kshrkLog);
    #else
    #endif

    /* Zero avlnodecore. */
    if (pMod)
        memset(&pMod->avlnodecore, 0, sizeof(pMod->avlnodecore));
}


/**
 * Log start seqence and entry function event.
 * @returns Dummy value. (0)
 * @param   hLogMod         Module Log handle.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   phSeq           Pointer to sequence handle (on the stack!).
 * @param   pszFuncProto    Function prototype string.
 *                          For C++ __FUNCTION__ is usually enough.
 * @param   fOptlink        Optlink (uses register (and not stack only) to pass arguments)
 * @param   ...             Parameters if optlink function.
 *                          Pointer to first parameter if not optlink.
 *                          NULL if no parameters.
 * @author  knut st. osmundsen (kosmunds@csc.com)
 */
int kLogStart(HKLOGMOD hLogMod, KLOGPOS_DECL, PHKLOGSEQ phSeq, const char *pszFuncProto, KBOOL fOptlink, ...)
{
    va_list         args;
    PKLOGMODDATA    pMod;

    /*
     * Get module data from handle.
     */
    pMod = klogGetMod(hLogMod); /* NB! might be NULL */

    /*
     * Set sequence handle.
     */
    phSeq = (PHKLOGSEQ)KALIGNUP(8, phSeq);
    phSeq->data.ulMagic = SEQ_MAGIC;
    phSeq->data.ulSeq   = SEQ_FLAGS | AtomicIncWrap(pMod ? &pMod->ulLastSeq : &ulGlobalLastSeq, SEQ_FLAGS);

    /*
     * Log the event.
     */
    va_start(args, fOptlink);
    klogEntry(phSeq, pMod, KLOGPOS_INT, KLOG_TYPE_START, pszFuncProto, args);
    va_end(args);

    return 0;                           /* Dummy */
}


/**
 * Log exit function event.
 * @param   hLogMod         Module Log handle.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   phSeq           Pointer to sequence handle. (optional)
 * @param   pszReturnType   Typedefinition of the return type. (optional)
 * @param   ...             The value of the return type.
 * @author  knut st. osmundsen (kosmunds@csc.com)
 */
void kLogStop(HKLOGMOD hLogMod, KLOGPOS_DECL, PHKLOGSEQ phSeq, const char *pszReturnType, ...)
{
    va_list         retarg;
    PKLOGMODDATA    pMod;

    /*
     * Get module data from handle.
     */
    pMod = klogGetMod(hLogMod);

    /* get phSeq if not passed in */
    if (!phSeq)
        phSeq = klogGetSeq(0);
    else
        phSeq = (PHKLOGSEQ)(void*)KALIGNUP(8, phSeq);

    /*
     * Log this event.
     */
    va_start(retarg, pszReturnType);
    klogExit(phSeq, pMod, KLOGPOS_INT, KLOG_TYPE_STOP, pszReturnType, retarg);
    va_end(retarg);

    /*
     * Destroy seqence handle.
     */
    if (phSeq)
    {
        phSeq->data.ulMagic = 0;
        phSeq->data.ulSeq = 0;
    }
}



/**
 * Log enter function event.
 * @returns Dummy value. (0)
 * @param   hLogMod         Module Log handle.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   pszFuncProto    Function prototype string.
 *                          For C++ __FUNCTION__ is usually enough.
 * @param   fOptlink        Optlink (uses register (and not stack only) to pass arguments)
 * @param   ...             Parameters if optlink function.
 *                          Pointer to first parameter if not optlink.
 *                          NULL if no parameters.
 * @author  knut st. osmundsen (kosmunds@csc.com)
 */
int kLogEntry(HKLOGMOD hLogMod, KLOGPOS_DECL, const char *pszFuncProto, KBOOL fOptlink, ...)
{
    va_list         args;
    PHKLOGSEQ       phSeq;
    PKLOGMODDATA    pMod;

    /*
     * Get module data from handle.
     */
    pMod = klogGetMod(hLogMod);

    /*
     * Get sequence handle.
     */
    phSeq = klogGetSeq(0);

    /*
     * Log this event.
     */
    va_start(args, fOptlink);
    if (!fOptlink)
    {
        va_list argtmp = va_arg(args, va_list);
        va_end(args);
        args = argtmp;
    }
    klogEntry(phSeq, pMod, KLOGPOS_INT, KLOG_TYPE_ENTRY, pszFuncProto, args);
    va_end(args);

    return 0;                           /* Dummy */
}


/**
 * Log exit function event.
 * @param   hLogMod         Module Log handle.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   pszReturnType   Typedefinition of the return type.
 * @param   ....            The value of the return type.
 * @author  knut st. osmundsen (kosmunds@csc.com)
 */
void kLogExit(HKLOGMOD hLogMod, KLOGPOS_DECL, const char *pszReturnType, ...)
{
    va_list         retarg;
    PHKLOGSEQ       phSeq;
    PKLOGMODDATA    pMod;

    /*
     * Get module data from handle.
     */
    pMod = klogGetMod(hLogMod);

    /*
     * Get sequence handle.
     */
    phSeq = klogGetSeq(0);

    /*
     * Log this event.
     */
    va_start(retarg, pszReturnType);
    klogExit(phSeq, pMod, KLOGPOS_INT, KLOG_TYPE_EXIT, pszReturnType, retarg);
    va_end(retarg);
}



/**
 * Generic logging function.
 * @param   hLogMod         Module Log handle.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   iType           Log type.
 * @param   ...             Optional parameters.
 * @author  knut st. osmundsen (kosmunds@csc.com)
 */
void kLog(HKLOGMOD hLogMod, KLOGPOS_DECL, int iType, ...)
{
    va_list         args;
    PHKLOGSEQ       phSeq;
    PKLOGMODDATA    pMod;

    /*
     * Get module data from handle.
     */
    pMod = klogGetMod(hLogMod);

    /*
     * Get sequence handle.
     */
    phSeq = klogGetSeq(0);

    /*
     * Log this event.
     */
    va_start(args, iType);
    klogv(phSeq, pMod, KLOGPOS_INT, iType, args);
    va_end(args);
}


/**
 * Generic logging function.
 * @param   hLogMod         Module Log handle.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   iType           Log type.
 * @param   args            Optional parameters.
 * @author  knut st. osmundsen (kosmunds@csc.com)
 * @remark  Performs va_end on args.!
 */
void kLogv(HKLOGMOD hLogMod, KLOGPOS_DECL, int iType, va_list args)
{
    PHKLOGSEQ       phSeq;
    PKLOGMODDATA    pMod;

    /*
     * Get module data from handle.
     */
    pMod = klogGetMod(hLogMod);

    /*
     * Get sequence handle.
     */
    phSeq = klogGetSeq(0);

    /*
     * Log this event.
     */
    klogv(phSeq, pMod, KLOGPOS_INT, iType, args);

    va_end(args);
}


/**
 * Hack to support kprintf and dprintf - part 1.
 * Will do the formatting of the passed in data to a buffer.
 * The buffer is returned and passed on to kLogfixPrintf2.
 * @returns Buffer index.
 * @param   pszFormat   Pointer to format string.
 * @param   ...         Optional parameters.
 */
int kLogFixPrintf1(const char *pszFormat, ...)
{
    va_list         args;
    int             iBuffer;
    int             cchBuffer;

    /*
     * Allocate buffer.
     */
    iBuffer = klogGetBuffer();

    /*
     * Format output.
     */
    va_start(args, pszFormat);
    cchBuffer = vsprintf(KLOG_GETBUFFERADDRESS(iBuffer), pszFormat, args);
    va_end(args);


    #ifdef DEBUG
    /*
     * Check for overflow.
     */
    if (cchBuffer >= KLOG_MAX_OVERFLOW - 42) /* - 42: common stuff. */
    {
        klog(NULL, NULL, KLOGPOS_EXT, KLOG_TYPE_LOGASSERT,
             "overflowed buffer %d with %d bytes",
             iBuffer,
             cchBuffer - KLOG_BUFFER_ENTRY_SIZE);
    }
    #endif

    return iBuffer;
}


/**
 * Hack to support kprintf and dprintf - part 2.
 * Receives a log event with a buffer index as parameter.
 * Will do logging of the content of the buffer and free the buffer.
 *
 * @param   hLogMod         Module Log handle.
 * @param   KLOGPOS_DECL    Source code position stuff.
 * @param   iType           Log event type. Usually KLOG_TYPE_PRINTF.
 * @param   iBuffer         Buffer index which contains the data which we're to
 *                          pass on to kLog.
 */
void kLogFixPrintf2(HKLOGMOD hLogMod, KLOGPOS_DECL, int iType, int iBuffer)
{
    /*
     * Log the print event.
     */
    kLog(hLogMod, KLOGPOS_INT, iType, KLOG_GETBUFFERADDRESS(iBuffer));

    /*
     * Free the buffer with the sprintf output.
     */
    klogFreeBuffer(iBuffer);
}

